# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from .leaf import Leaf


class NXdisk_chopper(Leaf):
    """Disk chopper information, can be read as a dict."""
